from __future__ import annotations
from copy import copy
import random
import json
from pathlib import Path
import numpy as np
from sklearn.base import BaseEstimator
from .config import hh_logger
from .hyperparam_info import known_hyperparams, known_models, hyperparam_def_schema

def add_known_model_def(def_dict: dict):
    """
    Adds the definition of the hyperparameters of a new model to the
    list of know hyperparameters and known models.

    The definition should be a dictionary that follows this schema:

    {
        'model': <Name>,
        <hyperparam_name>:
            Choose one of:
            - {"type": "real", "min": <number>, "max": <number>},
            - {"type": "integer", "min": <number>, "max": <number>},
            - {"type": "categorical", "values": [<number or string>]}
    }

    There can be any number of hyperparameters, even 0, they MUST correspond to the
    arguments used in the model constructor or you will get an error in the hyperparameter
    search step.
    
    :param def_dict: dictionary that defines the hyperparameters of a new model.
    """

    if hyperparam_def_schema.is_valid(def_dict):
        known_hyperparams.append(def_dict)
        known_models.append(def_dict["model"])
    else:
        hh_logger.error("The hyperparameter definition provided has an incorrect format.")

        # Since the schema is incorrect, validating the schema with raise an error with
        # detailed information about why it failed.
        hyperparam_def_schema.validate(def_dict)


def find_hyperparam_grid(model: BaseEstimator, n_grid_points: int = 10) -> dict:
    """
    Obtains a grid of hyperparameters to optimize for the model.

    :param model: List of model for which we want to find the hyperparameters.
    :param n_grid_points: Number of values to pick for each hyperparameter.
    :return: List with the hyperparameter definition for each of the model passed
    """

    hyperparam_def = find_hyperparam_def(model)
    if hyperparam_def:
        hyperparam_grid = construct_hyperparam_grid(hyperparam_def, n_grid_points)
        hh_logger.info(f"The hyperparameter grid for the model {type(model).__name__} was generated succesfully.")
    else:
        hyperparam_grid = {}

    return hyperparam_grid

def find_hyperparam_def(model: BaseEstimator) -> dict:
    """
    Obtains the definitions of the hyperparameters of each of the model listed.

    :param model: List of model for which we want to find the hyperparameters.
    :return: List with the hyperparameter definition for each of the model passed
    """

    model_name = type(model).__name__
    params = None

    if model_name in known_models:
        p_idx = known_models.index(model_name)
        params = copy(known_hyperparams[p_idx])
        params.pop("model", None)
    else:
        hh_logger.warning(f"The model {model_name} has not been found and no hyperparameter definitions can be used.")

    return params

def construct_hyperparam_grid(hyperparam_grid_def: dict, n_grid_points: int = 10) -> dict:
    """
    Generate a grid of hyperparameters from their definition.

    :param hyperparam_grid_def: Definition of the hyperparameters to be generated as a grid.
    :param n_grid_points: Number of values to pick for each hyperparameter.
    :return: List of hyperparameter grids to use in grid search.
    """

    keys = [k for k in hyperparam_grid_def.keys() if k != 'model']

    model_params = {}
    hh_logger.debug(f"Using hyperparameter definitions: {hyperparam_grid_def}")
    for k in keys:
        if hyperparam_grid_def[k]["type"] == "integer":
            model_params[k] = np.linspace(hyperparam_grid_def[k]["min"], hyperparam_grid_def[k]["max"], n_grid_points)

            # Convert linspace generated numbers to integers
            model_params[k] = np.unique(np.round(model_params[k])).astype(int)

        elif hyperparam_grid_def[k]["type"] == "real":
            model_params[k] = np.geomspace(max(hyperparam_grid_def[k]["min"], 1e-10), hyperparam_grid_def[k]["max"], n_grid_points)

        elif hyperparam_grid_def[k]["type"] == "categorical":
            values = hyperparam_grid_def[k]["values"]

            # Take at most n random non-repeating values
            model_params[k] = random.sample(values, k=min(len(values), n_grid_points))

            # Interpret 'None' as a python null value
            model_params[k] = [i if i != 'None' else None for i in model_params[k]]

    return model_params



import numpy as np
import seaborn as sns
from adjustText import adjust_text
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split
from .metric_alias import metric_alias


def _save_plot(filepath, display, dpi=200):
    if filepath is not None:
        plt.savefig(filepath, dpi=dpi)
        if not display:
            plt.close()
    else:
        if display:
            plt.show()


def plot_confusion_matrix(X, y, model, class_dict, title="", test_size=0.2, seed=0, filepath=None, display=True):
    """
    Plot confusion matrix for a given model.

    :param X: input observations
    :param y: target values
    :param model: model to evaluate
    :param class_dict: dictionary with class names (ex.: {0: "No", 1: "Yes"})
    :param title: title of the plot
    :param test_size: percentage of the dataset to use for testing
    :param seed: random seed
    :param filepath: path to save the plot
    :param display: whether to display the plot
    """

    # Setting up
    labels = [class_dict[f] for f in np.unique(y)]
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, test_size=test_size, random_state=seed)

    # Plotting
    sns.set_style("ticks")
    fig, axs = plt.subplots(2, 2, figsize=(10, 10))
    fig.patch.set_facecolor('lightgrey')

    for i, md in enumerate([model, DummyClassifier()]):
        md.fit(X_train, y_train)

        cm_train = confusion_matrix(y_train, md.predict(X_train))
        cm_test = confusion_matrix(y_test, md.predict(X_test))

        for j, (cm_name, cm, _X, _y) in enumerate([("Train", cm_train, X_train, y_train),
                                                   ("Test", cm_test, X_test, y_test)]):
            ax = axs[i, j]

            percentages = cm.astype('float') / cm.sum()
            annots = [[f'{v}\n\n({(p*100):.2f})' for v, p in zip(vs, ps)] for vs, ps in zip(cm, percentages)]
            cmap = sns.color_palette(["light:#7A7", "light:#77B"][j], as_cmap=True)

            sns.heatmap(cm, annot=annots, fmt='', square=True, xticklabels=labels, ax=ax,
                        yticklabels=labels, linewidths=1, cmap=cmap, cbar=False, linecolor='black')

            md_name = md.__class__.__name__
            acc = md.score(_X, _y)
            ax_title = f"{md_name} // {cm_name} data\n(N: {len(_y)}, Acc: {acc:.3f})"
            ax.set_title(ax_title, fontweight='bold')

            ax.set_xlabel("Predicted", fontweight='bold')
            ax.set_ylabel("Actual", fontweight='bold')

    plt.suptitle(title, fontweight='bold', fontsize=14)
    plt.tight_layout()

    _save_plot(filepath, display)

def plot_regression_pred(X, y, models, y_label="", title="", test_size=0.2,
                         metric=None, seed=0, filepath=None, display=True):
    """
    Plot the predictions of the regression model

    :param X: input observations
    :param y: target values
    :param models: list of models to evaluate
    :param y_label: name of the target variable
    :param title: title of the plot
    :param test_size: percentage of the dataset to use for testing
    :param metric: metric to use for evaluation
    :param seed: random seed
    :param filepath: path to save the plot
    :param display: whether to display the plot
    """

    # Preparing data
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=test_size, random_state=seed)

    _s = np.argsort(y_train)
    X_train, y_train = X_train[_s], y_train[_s]
    _s = np.argsort(y_test)
    X_test, y_test = X_test[_s], y_test[_s]

    _X = np.concatenate((X_train, X_test))
    _y = np.concatenate((y_train, y_test))

    # Plotting
    sns.set_style("ticks")

    fig = plt.figure(figsize=(10, 6))
    fig.patch.set_facecolor('lightgrey')

    plt.plot(range(len(_y)), _y, 'b', label="Data", linewidth=3)
    for md in models:
        md = md.fit(X_train, y_train)
        md_name = md.__class__.__name__

        res = metric(y_test, md.predict(X_test)) if metric is not None else ""
        metric_name = metric.__name__ if metric is not None else ""
        label = f"{md_name}" + (f"\n({metric_name}: {res:.2f})" if metric is not None else "")

        plt.plot(range(len(_y)), md.predict(_X), label=label)

    plt.axvline(len(y_train) - 0.5, color='k', linestyle='--', linewidth=3)
    plt.text(len(y_train) / 2, _y.max() * 0.9, "Train data", ha='center', va='center', fontsize=20)
    plt.text(len(y_train) + len(y_test) / 2, _y.max() * 0.9, "Test data", ha='center', va='center', fontsize=20)
    plt.gca().add_patch(Rectangle((len(y_train) - 0.5, _y.min()), len(y_test), (_y.max() - _y.min()),
                                  fill=True, alpha=0.1, color='b'))

    plt.title(title, fontweight='bold')
    plt.xlabel("Samples")
    plt.ylabel(y_label)
    plt.ylim(min(_y), max(_y))
    plt.xlim(0, len(_y))
    plt.legend()
    plt.grid(True)

    plt.tight_layout()

    _save_plot(filepath, display)

def plot_batch_results(df, metric_name, title="", filepath=None, display=True):
    """
    Plot the results of the batch evaluation

    :param df: results dataframe
    :param title: title of plot
    :param filepath: filepath to save plot
    :param display: whether to display the plot
    """

    # sns.set_context("paper")
    plt.figure(figsize=(6, 6))
    ax = plt.gca()

    y_label = f"Avg {metric_name} (Validation Test)"
    x_label = f"Avg {metric_name} (Validation Train)"
    baseline_result = df[df["Model"].str.startswith("Dummy")][y_label].values[0]

    sns.scatterplot(data=df, x=x_label, y=y_label, s=100, hue="Model", legend=False)
    ax.axhline(baseline_result, color="grey", linestyle='--', linewidth=2, zorder=-1)
    ax.set_facecolor("#eeeeee")

    texts = []
    for name, row in df.iterrows():
        text = plt.text(row[x_label] + 0.005, row[y_label] + 0.001, row["Model"], fontsize=12)
        texts.append(text)
    adjust_text(texts, arrowprops=dict(arrowstyle="-", color='k', lw=0.5), force_text=1.0, force_points=1.0)

    plt.title(title, fontweight='bold')
    plt.grid()

    plt.tight_layout()

    _save_plot(filepath, display)


def plot_multiple_datasets(df, metric_name, id_col="Code", title="", line_at_0=False,
                           higher_is_better=True, filepath=None, display=True):
    """
    Plot the results of the batch evaluation

    :param df: results dataframe
    :param metric_name: metric to plot
    :param id_col: column containing the ID of the dataset
    :param title: title of plot
    :param line_at_0: determines if a line is plotted at 0
    :param higher_is_better: determines if higher values are better
    :param filepath: filepath to save plot
    :param display: whether to display the plot
    """

    _df = df.sort_values(by=metric_name, ascending=(not higher_is_better))

    plt.figure(figsize=(10, 8))
    ax = plt.gca()

    sns.scatterplot(data=_df, x=id_col, y=metric_name, s=100, hue=id_col, legend=False)

    ax.set_xlabel("")
    ax.set_ylabel(metric_name)

    if line_at_0:
        ax.axhline(0, lw=4, color='k', zorder=-1)

    for tick in ax.get_xticklabels():
        tick.set_rotation(90)

    ax.grid()
    ax.set_facecolor("#eeeeee")

    plt.suptitle(title, fontweight='bold')
    plt.tight_layout()

    _save_plot(filepath, display)


from .classifier import HundredHammersClassifier
from .regressor import HundredHammersRegressor
from .metric_alias import metric_alias
from .config import hh_logger
from .hyperparameters import known_hyperparams, known_models, hyperparam_def_schema, add_known_model_def
from .plots import plot_batch_results, plot_confusion_matrix, plot_regression_pred

__version__ = "0.1.0"



"""
This module provides some alternative names for metrics
implemented in sklearn.
"""

metric_alias = {
    "ACC": "accuracy",
    "BACC": "balanced_accuracy",
    "PREC": "precision",
    "PRECW": "precision_weighted",
    "REC": "recall",
    "F1": "f1",
    "F1W": "f1_weighted",
    "ROC": "roc_auc",
    "LogLoss": "neg_log_loss",
    "MAE": "neg_mean_absolute_error",
    "RMSE": "neg_root_mean_squared_error",
    "MAPE": "neg_mean_absolute_percentage_error",
    "MSE": "neg_mean_squared_error",
    "R2": "r2",
}



# pylint: skip-file
import pandas as pd
import pytest
import logging
from sklearn.datasets import load_iris
import matplotlib.pyplot as plt

from hundred_hammers import hh_logger
from hundred_hammers import HundredHammersClassifier
from hundred_hammers.plots import plot_batch_results, plot_confusion_matrix
from hundred_hammers.model_zoo import KNeighborsClassifier

import warnings
from sklearn.exceptions import ConvergenceWarning

hh_logger.setLevel(logging.DEBUG)

def test_complete_default():
    data = load_iris()
    X = data.data
    y = data.target

    # Create the model
    hh = HundredHammersClassifier()

    # Evaluate the model
    df_results = hh.evaluate(X, y, optim_hyper=False)
    assert type(df_results) is pd.DataFrame

def test_complete_optim_hyperparams():
    data = load_iris()
    X = data.data
    y = data.target

    # Create the model
    hh = HundredHammersClassifier()

    # Evaluate the model
    df_results = hh.evaluate(X, y, optim_hyper=True, n_grid_points=4)
    assert type(df_results) is pd.DataFrame

def test_seed_strategy():
    data = load_iris()
    X = data.data
    y = data.target

    df_results_1 = HundredHammersClassifier(seed_strategy='sequential').evaluate(X, y, optim_hyper=False)
    df_results_2 = HundredHammersClassifier(seed_strategy='sequential').evaluate(X, y, optim_hyper=False)

    assert df_results_1.equals(df_results_2)

    df_results_3 = HundredHammersClassifier(seed_strategy='random').evaluate(X, y, optim_hyper=False)
    df_results_4 = HundredHammersClassifier(seed_strategy='random').evaluate(X, y, optim_hyper=False)

    assert not df_results_3.equals(df_results_4)


"""
This module provides an easy way of accessing all available machine learning models
and provides some default models to use for classification and regression tasks.
"""

from sklearn.dummy import DummyClassifier # pylint: disable=W0611
from sklearn.dummy import DummyRegressor # pylint: disable=W0611

from sklearn.isotonic import IsotonicRegression # pylint: disable=W0611

from sklearn.neighbors import KNeighborsClassifier # pylint: disable=W0611
from sklearn.neighbors import KNeighborsRegressor # pylint: disable=W0611

from sklearn.linear_model import LinearRegression # pylint: disable=W0611
from sklearn.linear_model import Ridge # pylint: disable=W0611
from sklearn.linear_model import RidgeCV # pylint: disable=W0611
from sklearn.linear_model import Lasso # pylint: disable=W0611
from sklearn.linear_model import LassoCV # pylint: disable=W0611
from sklearn.linear_model import ElasticNet # pylint: disable=W0611
from sklearn.linear_model import ElasticNetCV # pylint: disable=W0611
from sklearn.linear_model import LogisticRegression # pylint: disable=W0611
from sklearn.linear_model import RidgeClassifier # pylint: disable=W0611
from sklearn.linear_model import RidgeClassifierCV # pylint: disable=W0611
from sklearn.linear_model import SGDClassifier # pylint: disable=W0611
from sklearn.linear_model import Perceptron # pylint: disable=W0611
from sklearn.linear_model import PassiveAggressiveClassifier # pylint: disable=W0611
from sklearn.linear_model import PassiveAggressiveRegressor # pylint: disable=W0611

from sklearn.kernel_ridge import KernelRidge# pylint: disable=W0611

from sklearn.neural_network import MLPClassifier # pylint: disable=W0611
from sklearn.neural_network import MLPRegressor # pylint: disable=W0611

from sklearn.svm import LinearSVC # pylint: disable=W0611
from sklearn.svm import LinearSVR # pylint: disable=W0611
from sklearn.svm import SVC # pylint: disable=W0611
from sklearn.svm import SVR # pylint: disable=W0611

from sklearn.tree import DecisionTreeClassifier # pylint: disable=W0611
from sklearn.tree import DecisionTreeRegressor # pylint: disable=W0611
from sklearn.tree import ExtraTreeClassifier # pylint: disable=W0611
from sklearn.tree import ExtraTreeRegressor # pylint: disable=W0611

from sklearn.ensemble import AdaBoostClassifier # pylint: disable=W0611
from sklearn.ensemble import AdaBoostRegressor # pylint: disable=W0611
from sklearn.ensemble import GradientBoostingClassifier # pylint: disable=W0611
from sklearn.ensemble import GradientBoostingRegressor # pylint: disable=W0611
from sklearn.ensemble import RandomForestClassifier # pylint: disable=W0611
from sklearn.ensemble import RandomForestRegressor # pylint: disable=W0611
from sklearn.ensemble import ExtraTreesClassifier # pylint: disable=W0611
from sklearn.ensemble import BaggingClassifier # pylint: disable=W0611

from sklearn.naive_bayes import BernoulliNB # pylint: disable=W0611
from sklearn.naive_bayes import CategoricalNB # pylint: disable=W0611
from sklearn.naive_bayes import ComplementNB # pylint: disable=W0611
from sklearn.naive_bayes import GaussianNB # pylint: disable=W0611
from sklearn.naive_bayes import MultinomialNB # pylint: disable=W0611

from sklearn.discriminant_analysis import LinearDiscriminantAnalysis # pylint: disable=W0611
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis # pylint: disable=W0611

from sklearn.gaussian_process import GaussianProcessClassifier # pylint: disable=W0611
from sklearn.gaussian_process import GaussianProcessRegressor # pylint: disable=W0611

# XGBoost library
from xgboost import XGBRegressor # pylint: disable=W0611
from xgboost import XGBClassifier # pylint: disable=W0611

# scikit-elm library
from skelm import ELMRegressor # pylint: disable=W0611
from skelm import ELMClassifier # pylint: disable=W0611

DEFAULT_REGRESSION_METRICS = [
    "R2", "MSE", "MAE"
]

DEFAULT_CLASSIFICATION_METRICS = [
    "ACC", "F1W"
]

DEFAULT_REGRESSION_MODELS = [
    ("Dummy Mean", DummyRegressor(strategy="mean"), {}),
    ("Dummy Median", DummyRegressor(strategy="median"), {}),
    ("Linear Regression", LinearRegression(), {}),
    ("Decision Tree", DecisionTreeRegressor(), {}),
    ("SVR", SVR(), {}),
    ("Linear SVR", LinearSVR(dual="auto"), {}),
    ("Ridge", Ridge(),{}),
    ("Passive Aggressive", PassiveAggressiveRegressor(), {}),
    ("KNN", KNeighborsRegressor(), {}),
    ("Neural Network Regressor", MLPRegressor(), {}),
    ("Gaussian Process", GaussianProcessRegressor(), {}),
    ("Random Forest", RandomForestRegressor(), {}),
    ("AdaBoost", AdaBoostRegressor(), {}),
    ("Gradient Boosting", GradientBoostingRegressor(), {})
]

DEFAULT_CLASSIFICATION_MODELS = [
    ("Dummy", DummyClassifier(strategy="most_frequent"), {}),
    ("Decision Tree", DecisionTreeClassifier(random_state=0), {}),
    ('SVC', SVC(gamma='auto', kernel='rbf'), {}),
    ('Linear SVC', LinearSVC(random_state=0, tol=1e-5, dual="auto"), {}),
    ('Perceptron', Perceptron(random_state=0), {}),
    ('Logistic Regression', LogisticRegression(random_state=0, solver='lbfgs', multi_class='auto'), {}),
    ('Ridge Classifier', RidgeClassifier(random_state=0), {}),
    ('SGD Classifier', SGDClassifier(random_state=0), {}),
    ('Passive Aggressive Classifier', PassiveAggressiveClassifier(random_state=0), {}),
    ('K Neighbors Classifier', KNeighborsClassifier(), {}),
    ('Neural Network Classifier', MLPClassifier(random_state=0), {}),
    ('Gaussian Process Classifier', GaussianProcessClassifier(random_state=0), {}),
    ('Random Forest Classifier', RandomForestClassifier(random_state=0), {}),
    ('AdaBoost Classifier', AdaBoostClassifier(random_state=0), {}),
    ('Gradient Boosting Classifier', GradientBoostingClassifier(random_state=0), {}),
    ('Extra Trees Classifier', ExtraTreesClassifier(random_state=0), {}),
    ('Bagging Classifier', BaggingClassifier(random_state=0), {})
]



"""
Global configuration for the library.
"""

import os
import logging

os.environ['PYTHONWARNINGS']='ignore'

logging.basicConfig(format='[%(levelname)s] %(asctime)s: %(message)s')
hh_logger = logging.getLogger(__name__)
hh_logger.setLevel(logging.INFO)



from schema import Schema, Or, Optional, Regex

hyperparam_def_schema = Schema({
    'model': str,
    Optional(Regex('.*')): Or(
        {"type": "real", "min": Or(float, int), "max": Or(float, int)},
        {"type": "integer", "min": int, "max": int},
        {"type": "categorical", "values": [lambda x: True]}, # Accept a list of any type
    )
})

known_hyperparams = [
    {
        "model": "DummyClassifier",
        "strategy": {
            "type": "categorical",
            "values": [
                "most_frequent",
                "prior",
                "stratified",
                "uniform"
            ]
        }
    },

    {
        "model": "DummyRegressor",
        "strategy": {
            "type": "categorical",
            "values": [
                "mean",
                "median"
            ]
        }
    },

    {
        "model": "IsotonicRegression"
    },

    {
        "model": "KNeighborsClassifier",
        "n_neighbors": {
            "type": "integer",
            "min": 1,
            "max": 500
        },
        "metric": {
            "type": "categorical",
            "values": [
                "manhattan",
                "euclidean",
                "cosine"
            ]
        }
    },

    {
        "model": "KNeighborsRegressor",
        "n_neighbors": {
            "type": "integer",
            "min": 1,
            "max": 500
        },
        "metric": {
            "type": "categorical",
            "values": [
                "manhattan",
                "euclidean",
                "cosine"
            ]
        }
    },

    {
        "model": "LinearRegression"
    },

    {
        "model": "Ridge",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        }
    },

    {
        "model": "Lasso",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        }
    },

    {
        "model": "ElasticNet",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "l1_ratio": {
            "type": "real",
            "min": 0,
            "max": 1
        }
    },

    {
        "model": "LogisticRegression"
    },

    {
        "model": "RidgeClassifier",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        }
    },

    {
        "model": "SGDClassifier",
        "penalty": {
            "type": "categorical",
            "values": [
                "l1",
                "l2",
                "elasticnet",
                "None"
            ]
        },
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "l1_ratio": {
            "type": "real",
            "min": 0,
            "max": 1
        }
    },

    {
        "model": "Perceptron",
        "penalty": {
            "type": "categorical",
            "values": [
                "l1",
                "l2",
                "elasticnet",
                "None"
            ]
        },
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "l1_ratio": {
            "type": "real",
            "min": 0,
            "max": 1
        }
    },

    {
        "model": "PassiveAggressiveClassifier",
        "C": {
            "type": "real",
            "min": 0,
            "max": 100
        }   
    },

    {
        "model": "KernelRidge",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "kernel": {
            "type": "categorical",
            "values": [
                "linear",
                "poly",
                "rbf"
            ]
        }
    },

    {
        "model": "SVC",
        "C": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "kernel": {
            "type": "categorical",
            "values": [
                "linear",
                "poly",
                "rbf"
            ]
        },
        "degree": {
            "type": "integer",
            "min": 1,
            "max": 6
        }
    },

    {
        "model": "SVR",
        "C": {
            "type": "real",
            "min": 0,
            "max": 100
        },
        "kernel": {
            "type": "categorical",
            "values": [
                "linear",
                "poly",
                "rbf"
            ]
        },
        "degree": {
            "type": "integer",
            "min": 1,
            "max": 6
        }
    },

    {
        "model": "DecisionTreeClassifier",
        "criterion": {
            "type": "categorical",
            "values": [
                "gini",
                "entropy",
                "log_loss"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 20
        }
    },

    {
        "model": "DecisionTreeRegressor",
        "criterion": {
            "type": "categorical",
            "values": [
                "friedman_mse",
                "squared_error",
                "absolute_error",
                "poisson"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "AdaBoostClassifier",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        }
    },

    {
        "model": "AdaBoostRegressor",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        }
    },

    {
        "model": "GradientBoostingClassifier",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        },
        "criterion": {
            "type": "categorical",
            "values": [
                "friedman_mse",
                "squared_error"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "GradientBoostingRegressor",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        },
        "criterion": {
            "type": "categorical",
            "values": [
                "friedman_mse",
                "squared_error"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "RandomForestClassifier",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        },
        "criterion": {
            "type": "categorical",
            "values": [
                "gini",
                "entropy",
                "log_loss"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "RandomForestRegressor",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        },
        "criterion": {
            "type": "categorical",
            "values": [
                "friedman_mse",
                "squared_error",
                "absolute_error",
                "poisson"
            ]
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "BernoulliNB",
        "alpha": {
            "type": "real",
            "min": 0,
            "max": 1
        }
    },

    {
        "model": "GaussianNB",
        "var_smoothing": {
            "type": "real",
            "min": 0,
            "max": 1
        }
    },

    {
        "model": "LinearDiscriminantAnalysis",
        "solver": {
            "type": "categorical",
            "values": [
                "svd",
                "lsqr",
                "eigen"
            ]
        }
    },

    {
        "model": "QuadraticDiscriminantAnalysis"
    },

    {
        "model": "XGBRegressor",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 400
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "XGBClassifier",
        "n_estimators": {
            "type": "integer",
            "min": 1,
            "max": 100
        },
        "max_depth": {
            "type": "integer",
            "min": 1,
            "max": 30
        }
    },

    {
        "model": "ELMRegressor",
        "n_neurons": {
            "type": "integer",
            "min": 1,
            "max": 400
        }
    },

    {
        "model": "ELMClassifier",
        "n_neurons": {
            "type": "integer",
            "min": 1,
            "max": 400
        }
    }
]

known_models = [k["model"] for k in known_hyperparams]


# pylint: skip-file

import pandas as pd
import logging
from sklearn.datasets import load_diabetes
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt

from hundred_hammers import hh_logger
from hundred_hammers import HundredHammersRegressor
from hundred_hammers.plots import plot_batch_results, plot_regression_pred
from hundred_hammers.model_zoo import (
    DecisionTreeRegressor,
    LinearRegression,
    DummyRegressor
)

import warnings
from sklearn.exceptions import ConvergenceWarning

hh_logger.setLevel(logging.DEBUG)

def test_complete_default():
    data = load_diabetes()
    X = data.data
    y = data.target

    # Create the model
    hh = HundredHammersRegressor()

    # Evaluate the model
    df_results = hh.evaluate(X, y, optim_hyper=False)
    assert type(df_results) is pd.DataFrame

def test_complete_optim_hyperparams():
    data = load_diabetes()
    X = data.data
    y = data.target

    # Create the model
    hh = HundredHammersRegressor()

    # Evaluate the model
    df_results = hh.evaluate(X, y, optim_hyper=True, n_grid_points=4)
    assert type(df_results) is pd.DataFrame



from __future__ import annotations
from typing import Tuple, List, Iterable
import warnings
import random
from copy import deepcopy, copy
import pandas as pd
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split, KFold
from sklearn.metrics import get_scorer
from sklearn.model_selection import GridSearchCV
from .config import hh_logger
from .metric_alias import metric_alias
from .hyperparameters import find_hyperparam_grid
from rich.progress import Progress, TaskID, TimeElapsedColumn, SpinnerColumn, TextColumn, BarColumn, TaskProgressColumn, \
    TimeRemainingColumn, MofNCompleteColumn


def _process_metric(metric: str | callable) -> Tuple[str, callable, dict]:
    result = None

    if isinstance(metric, str):
        # Metric given by its name
        metric_fn_name = metric
        if metric in metric_alias:
            metric_fn_name = metric_alias[metric]

        scorer = get_scorer(metric_fn_name)
        result = (metric, scorer._score_func, scorer._kwargs)
    else:
        # Metric given as a lambda function
        result = (metric.__name__, metric, {})

    return result


class HundredHammersBase:
    """
    Base HundredHammers class. 
    Implements methods for automatic machine learning like evaluating a list of models
    and performing hyperparameter optimization.

    :param models: List of models to evaluate.
    :param metrics: Metrics to use to evaluate the models.
    :param eval_metric: Target metric to use in hyperparameter optimization.
    :param test_size: Percentage of the dataset to use for testing.
    :param n_folds: Number of Cross Validation folds.
    :param n_folds_tune: Number of Cross Validation folds in grid search.
    :param n_evals: Number of times to repeat the training of the models.
    :param seed_strategy: Strategy used to generate the seeds for the different evaluations ('sequential' or 'random')
    """

    def __init__(self, models: Iterable[Tuple[str, BaseEstimator, dict]] = None,
                 metrics: Iterable[str | callable] = None, eval_metric: str | callable = None,
                 test_size: float = 0.2, n_folds: int = 5, n_folds_tune: int = 5, n_evals: int = 10,
                 show_progress_bar: bool = True, seed_strategy: str = 'sequential'):
        self.models = models
        self.metrics = [_process_metric(metric) for metric in metrics]

        if eval_metric is None:
            self.eval_metric = self.metrics[0]
        else:
            self.eval_metric = _process_metric(eval_metric)

        self.test_size = test_size
        self.n_folds = n_folds
        self.n_folds_tune = n_folds_tune
        self.n_evals = n_evals
        self.show_progress_bar = show_progress_bar
        self.seed_strategy = seed_strategy
        self._report = pd.DataFrame()
        self._best_params = []
        self._trained_models = self.models

    @property
    def report(self) -> pd.DataFrame:
        """
        Pandas dataframe reflecting the results of the last evaluation of the models.

        :return: Dataframe with the performance of each of the models.
        """

        if self._report.empty:
            hh_logger.warning("No reports available. "
                              "Use the `evaluate` method to generate a report.")

        return self._report

    @property
    def best_params(self) -> List[Tuple[str, dict]]:
        """
        List of the best hyperparameters found for each model.

        :return: List of the best hyperparameters obtained for each model. 
        """

        if not self._best_params:
            hh_logger.warning("No available hyperparameters. "
                              "Hyperparameter optimization not performed.")

        model_names = [m_tup[0] for m_tup in self.models]

        return list(zip(model_names, self._best_params))

    @property
    def trained_models(self) -> Iterable[tuple[str, BaseEstimator, dict]]:
        """
        Get the trained models.

        :return: A list of models in the form of tuples (name, model, hyperparameters).
        """

        if self._report.empty:
            hh_logger.warning("The models were not trained, returning untrained models. "
                              "Use the `evaluate` method to train them.")

        return self._trained_models

    def _calc_metrics(self, y_true: np.ndarray, y_pred: np.ndarray) -> List[float]:
        """
        Calculate metrics for a given model.

        :param y_true: True values.
        :param y_pred: Predicted values.
        :return: A list with the results for each metric.
        """

        return [metric_fn(y_true, y_pred, **metric_params) for _, metric_fn, metric_params in self.metrics]

    def evaluate(self, X: np.ndarray, y: np.ndarray, optim_hyper: bool = True,
                 n_grid_points: int = 10) -> pd.DataFrame:
        """
        Train every model to obtain its performance.

        :param X: Input data.
        :param y: Target data.
        :param optim_hyper: Whether to optimize the hyperparameters of the models.
        :param n_grid_points: Number of points to take for each hyperparameter in grid search.
        :return: Dataframe with the performance of each of the models.
        """

        if optim_hyper:
            new_models = self.tune_models(X, y, n_grid_points)
        else:
            new_models = deepcopy(self.models)

        report, trained_models = self._evaluate_models(X, y, new_models)

        self._report = report
        self._trained_models = [(m_name, tmodel, param_grid) for (m_name, _, param_grid), tmodel in zip(self.models, trained_models)]

        return report

    def optimize_hyperparams(self, X: np.ndarray, y: np.ndarray, n_grid_points: int = 10) -> List[dict]:
        """
        Obtain the best set of parameters for each of the models.

        :param X: Input data.
        :param y: Target data.
        :param n_grid_points: Number of points to take for each hyperparameter in grid search.
        :return: List of the best hyperparameters obtained for each model. 
        """

        self._best_params = []

        with self._make_progress_bar() as progress:
            task = progress.add_task("[bold]Optimizing hyperparameters...", total=len(self.models))

            for name, model, param_grid in self.models:
                best_params_model = self._optimize_model_hyperparams(X, y, model, param_grid, n_grid_points)
                self._best_params.append(best_params_model)

                progress.update(task, advance=1, description=f"[bold]Tuning:[/bold] {name}")

        return self._best_params

    def tune_models(self, X: np.ndarray, y: np.ndarray,
                    n_grid_points: int = 10) -> List[Tuple[str, BaseEstimator, dict]]:
        """
        Tune a model using cross-validation.

        :param X: Input observations.
        :param y: Target values.
        :param n_grid_points: Number of points to take for each hyperparameter in grid search.
        :return: The tuned model.
        """

        best_param_list = self.optimize_hyperparams(X, y, n_grid_points)

        new_models = []
        for (model_name, model, model_param_grid), best_params in zip(self.models, best_param_list):
            # change the parameters without overwriting the model
            configured_model = copy(model).set_params(**best_params)

            new_models.append((model_name, configured_model, model_param_grid))

        return new_models

    def _evaluate_models(self, X: np.ndarray, y: np.ndarray,
                         models: Iterable[Tuple[str, BaseEstimator, dict]]) -> Tuple[pd.DataFrame, list[BaseEstimator]]:
        """
        Evaluate all models on a given dataset with their default hyperparameters.
        
        :param X: Input observations.
        :param y: Target values.
        :return: A DataFrame with the results.
        """

        data = []
        trained_models = []

        with self._make_progress_bar() as progress:
            for i, (name, model, _) in enumerate(models):
                hh_logger.info(f"Running model [{i+1}/{len(models)}]: {name}")
                task = progress.add_task(f"[bold]Evaluating:[/bold] {name} [bright_black][{i+1}/{len(models)}][/bright_black]", total=self.n_evals)

                res, new_model = self._evaluate_model_cv_multiple_seeds(X, y, model, n_evals=self.n_evals,
                                                                        progress=progress, task=task)
                trained_models.append(new_model)

                val = {"Model": name}
                for i, (metric_name, _, _) in enumerate(self.metrics):
                    for j, data_name in enumerate(["Validation Train", "Validation Test", "Train", "Test"]):
                        val[f"Avg {metric_name} ({data_name})"] = np.mean([m[i] for m in res[j]])
                        val[f"Std {metric_name} ({data_name})"] = np.std([m[i] for m in res[j]])

                data.append(val)

        return pd.DataFrame(data), trained_models

    def _evaluate_model_cv_multiple_seeds(self, X: np.ndarray, y: np.ndarray,
                                          model: BaseEstimator, n_evals: int = 10,
                                          progress: Progress = None,
                                          task: TaskID = None) -> Tuple[list[list[float]], BaseEstimator]:
        """
        Evaluate a model multiple times, with a different seed every time.

        :param X: Input observations.
        :param y: Target values.
        :param model: Model to evaluate.
        :param n_evals: Number of times to train the model (each iteration uses a different seed).
        :return: A tuple with the results for validation train, validation test, train and test.
        """

        results_val_train, results_val_test = [], []
        results_train, results_test = [], []

        if self.seed_strategy == "random":
            seeds = np.random.randint(0, 1000000, n_evals)
        elif self.seed_strategy == "sequential":
            seeds = range(0, n_evals)
        else:
            raise ValueError(f"Unknown seed strategy: {self.seed_strategy}")

        # take `n_evals` different seeds
        for i, seed in enumerate(seeds):
            hh_logger.debug(f"Iteration [{i+1}/{n_evals-1}]")
            res = self._evaluate_model_cv(X, y, model, seed=seed)

            results_val_train += res[0]
            results_val_test += res[1]
            results_train.append(res[2])
            results_test.append(res[3])

            if task is not None:
                progress.update(task, advance=1)

        # Take the model trained with the last seed
        trained_model = res[4]

        results = [results_val_train, results_val_test, results_train, results_test]

        model_info = "Metrics:\n"
        for i, (metric_name, _, _) in enumerate(self.metrics):
            model_info += f"{i}: {metric_name}\n"
            for j, data_name in enumerate(["Validation Train", "Validation Test", "Train", "Test"]):
                avg_res = np.mean([m[i] for m in results[j]])
                std_res = np.std([m[i] for m in results[j]])
                model_info += f"\t{data_name}: {avg_res:.3f} ± {std_res:.3f}\n"
            model_info += "\n"
        hh_logger.info(model_info)

        return results, trained_model

    def _evaluate_model_cv(self, X: np.ndarray, y: np.ndarray, model: BaseEstimator,
                           seed: int = 0) -> tuple[list[list[float]], list[list[float]], list[float], list[float], BaseEstimator]:
        """
        Evaluate a model on a given dataset.

        :param X: Input observations.
        :param y: Target values.
        :param model: Model to evaluate.
        :param seed: Random seed.
        :return: A tuple with the results for validation train, validation test, train and test.
        """

        if hasattr(model, 'random_state'):
            model.random_state = seed

        kf = KFold(n_splits=self.n_folds, shuffle=True, random_state=seed)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=self.test_size, random_state=seed)

        results_val_train, results_val_test = [], []

        for split_idx, (train_index, test_index) in enumerate(kf.split(X_train, y_train)):
            val_model = copy(model)
            hh_logger.debug(f"Split [{split_idx}/{self.n_folds}]")

            X_val_train, X_val_test = X_train[train_index], X_train[test_index]
            y_val_train, y_val_test = y_train[train_index], y_train[test_index]

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                val_model.fit(X_val_train, y_val_train)

            result_val_train = self._calc_metrics(val_model.predict(X_val_train), y_val_train)
            result_val_test = self._calc_metrics(val_model.predict(X_val_test), y_val_test)

            results_val_train.append(result_val_train)
            results_val_test.append(result_val_test)

        trained_model = copy(model)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            trained_model.fit(X_train, y_train)

        result_train = self._calc_metrics(trained_model.predict(X_train), y_train)
        result_test = self._calc_metrics(trained_model.predict(X_test), y_test)

        return results_val_train, results_val_test, result_train, result_test, trained_model

    def _optimize_model_hyperparams(self, X: np.ndarray, y: np.ndarray,
                                    model: BaseEstimator, param_grid: dict = None,
                                    n_grid_points: int = 10) -> dict:
        """
        Optimize the hyperparameters of a model.
        
        :param X: Input data.
        :param y: Target data.
        :param model: Model to optimize.
        :param param_grid: Predefined hyperparameter grid.
        :param n_grid_points: Number of points to take for each hyperparameter in grid search.
        :return: The best hyperparameters found for the model.
        """

        if not param_grid:
            hh_logger.info(f"No specified hyperparameter grid for {type(model).__name__}."
                           " Generating hyperparameter grid.")
            param_grid = find_hyperparam_grid(model, n_grid_points)

        eval_metric = lambda y_true, y_pred: self.eval_metric[1](y_true, y_pred,
                                                                 **self.eval_metric[2])
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            grid_search_model = GridSearchCV(model, param_grid, scoring=eval_metric,
                                             n_jobs=-1, cv=self.n_folds_tune)
            grid_search_model.fit(X, y)

        results = pd.DataFrame(grid_search_model.cv_results_)
        results.dropna()
        best_params_df = results[results["rank_test_score"] == results["rank_test_score"].min()]
        best_params = best_params_df.head(1)['params'][0]

        return best_params

    def _make_progress_bar(self):
        return Progress(
            BarColumn(),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            MofNCompleteColumn(),
            TextColumn("[progress.description]{task.description}"),
            disable=(not self.show_progress_bar)
        )


from copy import copy, deepcopy
from hundred_hammers.base import HundredHammersBase
from hundred_hammers.model_zoo import DEFAULT_REGRESSION_MODELS, DEFAULT_REGRESSION_METRICS


class HundredHammersRegressor(HundredHammersBase):
    """
    HundredHammers class specialized in regression models.
    Implements methods for automatic machine learning like evaluating a list of models
    and performing hyperparameter optimization.

    :param models: List of models to evaluate (has a default list of models)
    :param metrics: Metrics to use to evaluate the models (has a default list of metrics)
    :param eval_metric: Target metric to use in hyperparameter optimization (default is the first metric in metrics)
    :param test_size: Percentage of the dataset to use for testing (default 0.2)
    :param n_folds: Number of Cross Validation folds (default 5)
    :param n_folds_tune: Number of Cross Validation folds in grid search (default 5)
    :param n_evals: Number of times to repeat the training of the models (default 10)
    :param show_progress_bar: Show progress bar in the evaluation (default False)
    :param seed_strategy: Strategy used to generate the seeds for the different evaluations ('sequential' or 'random')
    """

    def __init__(self, models=None, metrics=None, eval_metric=None, test_size=0.2,
                 n_folds=5, n_folds_tune=5, n_evals=10, show_progress_bar=False, seed_strategy='sequential'):
        if models is None:
            models = deepcopy(DEFAULT_REGRESSION_MODELS)

        if metrics is None:
            metrics = copy(DEFAULT_REGRESSION_METRICS)

        super().__init__(models, metrics, eval_metric, test_size, n_folds, n_folds_tune,
                         n_evals, show_progress_bar, seed_strategy)


from setuptools import setup

setup(name='hundred_hammers',
      version='0.1.1',
      description='Batch execution of ML models to a dataset',
      url='http://github.com/vgarciasc/hundred_hammers',
      author='Vinicius Garcia',
      author_email='vinizinho@vinizinho.net',
      license='MIT',
      packages=['hundred_hammers'],
      zip_safe=False)


# pylint: skip-file

import pytest
from hundred_hammers import known_hyperparams, known_models, hyperparam_def_schema, add_known_model_def
from schema import SchemaError

def test_verify_predef_params():
    for known_model_hyperparams in known_hyperparams:
        hyperparam_def_schema.validate(known_model_hyperparams)

@pytest.mark.parametrize("new_hyperparams", [
    {'model': 'Ex1', 'param1': {'type': 'real', 'min': 0, 'max': 10.24}},
    {'model': 'Ex2', 'param1': {'type': 'integer', 'min': 0, 'max': 5}},
    {'model': 'Ex3', 'param1': {'type': 'categorical', 'values': ["manhattan", "euclidean"]}},
    {'model': 'Ex4', 'param1': {'type': 'categorical', 'values': [0, 2, 6]}},
    {'model': 'Ex5', 'param1': {'type': 'categorical', 'values': [(10,100,10), (2,10,1), (2,3,3)]}},
    {'model': 'Ex6', 'param1': {'type': 'real', 'min': 0, 'max': 10.24},
                     'param2': {'type': 'categorical', 'values': ["manhattan", "euclidean"]}},
    {'model': 'Ex7', 'param1': {'type': 'integer', 'min': 0, 'max': 10},
                     'param2': {'type': 'real', 'min': 0, 'max': 10.24},
                     'param3': {'type': 'categorical', 'values': ["manhattan", "euclidean"]}},
    {'model': 'Ex8'}
])
def test_add_to_known_hyperparams(new_hyperparams):
    add_known_model_def(new_hyperparams)
    assert new_hyperparams in known_hyperparams
    assert new_hyperparams['model'] in known_models

@pytest.mark.parametrize("new_hyperparams", [
    {'model': 'Bad1', 'param1': {'type': 'real', 'min': 0}},
    {'model': 'Bad1', 'param1': {'type': 'real', 'max': 0}},
    {'model': 'Bad2', 'param1': {'type': 'integer'}},
    {'model': 'Bad3', 'param1': {'type': 'categorical'}},
    {'model': 'Bad4', 'param1': {'type': 'categorical', 'min': 0, 'max': 5}},
    {'param1': {'type': 'categorical', 'values': [0, 2, 6]}}
])
def test_add_to_known_hyperparams_bad(new_hyperparams):
    with pytest.raises(SchemaError):
        add_known_model_def(new_hyperparams)
    assert new_hyperparams not in known_hyperparams
    assert new_hyperparams['model'] not in known_models if 'model' in new_hyperparams else True



import json
import numpy as np

class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, o):
        if isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        elif isinstance(o, np.ndarray):
            return o.tolist()
        return json.JSONEncoder.default(self, o)
